//go:generate go mod vendor
//go:generate protoc --go_out=plugins=grpc,paths=source_relative:. --proto_path=. --proto_path=vendor/gitlab.mashroom.com/mashroom/backend mashroom_document.proto
//go:generate rm -r vendor
//go:generate go fmt ./...
//go:generate go fix ./...
//go:generate goimports -w ./
//go:generate go mod tidy

// Package document - api definition
package document // import "gitlab.mashroom.com/mashroom/backend/document/api"

import (
	"bytes"
	"context"
	"crypto/sha256"
	"encoding/base64"
	"errors"
	"hash"
	"io"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	_ "gitlab.mashroom.com/mashroom/backend/common/api/protobuf" // required vendor
	"google.golang.org/protobuf/proto"
)

// ErrDigestMismatch - error returned when sha hash is mismatched
var ErrDigestMismatch = grpcerr.New(grpcerr.InvalidState, "digest mismatch")

// ChunkSize - default chunk size to transfer
var ChunkSize = 1024 * 1024 * 2

// DownloaderClient local interface abstraction
type DownloaderClient interface {
	Recv() (*FileEvent, error)
}

// UploadSender local interface abstraction
type UploadSender interface {
	Send(*FileEvent) error
}

// UploaderClient local interface abstraction
type UploaderClient interface {
	UploadSender
	CloseAndRecv() (*FileUploadResponse, error)
}

// UploaderWriter implemns io.Writer for uploder client
type UploaderWriter struct {
	Client UploadSender
}

// Write write bytes to document service
func (w *UploaderWriter) Write(bs []byte) (int, error) {
	err := w.Client.Send(&FileEvent{
		Data: &FileData{
			Data: bs,
		},
	})
	if err != nil {
		return 0, err
	}

	return len(bs), nil
}

// DownloaderReader implemnts io.Reader for downloader client
type DownloaderReader struct {
	Client    DownloaderClient
	remainder []byte
	hasher    hash.Hash
	file      *File
	err       error
}

// Sha256Hash returns resulting sha256 hash of downloaded file
func (r *DownloaderReader) Sha256Hash() []byte {
	return r.hasher.Sum(nil)
}

// File returns file metadata
func (r *DownloaderReader) File() *File {
	return r.file
}

// Read reads bytes from remote client
func (r *DownloaderReader) Read(bs []byte) (n int, err error) {
	var ev *FileEvent

	if len(r.remainder) == 0 {
		ev, err = r.Client.Recv()

		if errors.Is(err, io.EOF) {
			r.err, err = err, nil
		}

		if err != nil {
			return
		}
	}

	if ev.GetInit() != nil {
		r.remainder = nil
		r.hasher = sha256.New()
		r.file = ev.GetInit().GetFile()

		return 0, nil
	}

	if ev.GetData() != nil {
		r.remainder = ev.GetData().GetData()
		_, _ = r.hasher.Write(r.remainder)
	}

	n = copy(bs, r.remainder)
	r.remainder = r.remainder[n:]

	if len(r.remainder) == 0 && r.err != nil {
		return 0, r.err
	}

	return
}

func initFileUpload(name, mime string, reader io.ReadSeeker) (init *FileInit, err error) {
	_, err = reader.Seek(0, io.SeekStart)
	if err != nil {
		return nil, err
	}

	hasher := sha256.New()

	size, err := io.Copy(hasher, reader)
	if err != nil {
		return nil, err
	}

	hash := hasher.Sum(nil)

	_, err = reader.Seek(0, io.SeekStart)
	if err != nil {
		return nil, err
	}

	return &FileInit{
		File: &File{
			Name:   name,
			Mime:   mime,
			Size:   size,
			Sha256: hash,
		},
	}, nil
}

// UploadFile uploads file to document service
func UploadFile(
	ctx context.Context,
	client DocumentServiceClient,
	name, mime string,
	reader io.ReadSeeker,
) (file *File, err error) {
	buf := make([]byte, ChunkSize)

	uploader, err := client.FileUpload(ctx)
	if err != nil {
		return nil, err
	}

	init, err := initFileUpload(name, mime, reader)
	if err != nil {
		return nil, err
	}

	err = uploader.Send(&FileEvent{
		Init: init,
	})
	if err != nil {
		return nil, err
	}

	return UploadFileBuffer(ctx, uploader, reader, buf)
}

// UpdateFile uploads file to document service
func UpdateFile(
	ctx context.Context,
	client DocumentServiceClient,
	name, mime string,
	reader io.ReadSeeker,
) (file *File, err error) {
	init, err := initFileUpload(name, mime, reader)
	if err != nil {
		return nil, err
	}

	resp, err := client.FileUpdate(ctx, &FileUpdateRequest{
		File: init.GetFile(),
	})
	if err != nil {
		return nil, err
	}

	return resp.GetFile(), nil
}

// PlainReader ensures no WriteTo/ReadFrom
type PlainReader struct {
	R io.Reader
}

func (pr PlainReader) Read(bs []byte) (int, error) {
	return pr.R.Read(bs)
}

// PlainWriter ensures no WriteTo/ReadFrom
type PlainWriter struct {
	W io.Writer
}

func (pw PlainWriter) Write(bs []byte) (int, error) {
	return pw.W.Write(bs)
}

// UploadFileBuffer transfers bytes of file after init exchange using provied buffer
func UploadFileBuffer(
	ctx context.Context,
	uploader UploaderClient,
	reader io.Reader,
	buf []byte,
) (file *File, err error) {
	writer := &UploaderWriter{
		Client: uploader,
	}

	_, err = io.CopyBuffer(writer, PlainReader{R: reader}, buf)
	if err != nil {
		return nil, err
	}

	resp, err := uploader.CloseAndRecv()
	if err != nil {
		return nil, err
	}

	return resp.GetFile(), nil
}

// DownloadFile downloads file with given id and type from documents service
func DownloadFile(
	ctx context.Context,
	client DocumentServiceClient,
	id string,
	size ImageSize,
	writer io.Writer,
) (file *File, err error) {
	buf := make([]byte, ChunkSize)

	downloader, err := client.FileDownload(ctx, &FileDownloadRequest{
		Id:   id,
		Size: size,
	})
	if err != nil {
		return nil, err
	}

	return DownloadFileBuffer(ctx, downloader, writer, buf)
}

// DownloadBundleFile downloads file bunlde of given types from documents service
func DownloadBundleFile(
	ctx context.Context,
	client DocumentServiceClient,
	user, property string,
	types []DocumentType,
	writer io.Writer,
) (file *File, err error) {
	buf := make([]byte, ChunkSize)

	downloader, err := client.BundleDownload(ctx, &BundleDownloadRequest{
		User:     user,
		Property: property,
		Types:    types,
	})
	if err != nil {
		return nil, err
	}

	return DownloadFileBuffer(ctx, downloader, writer, buf)
}

// DownloadFileBuffer receives bytes from remote client after init exchange using provided buffer
func DownloadFileBuffer(
	ctx context.Context,
	downloader DownloaderClient,
	writer io.Writer,
	buf []byte,
) (file *File, err error) {
	reader := &DownloaderReader{
		Client: downloader,
	}

	_, err = io.CopyBuffer(PlainWriter{W: writer}, reader, buf)
	if err != nil && !errors.Is(err, io.EOF) {
		return nil, err
	}

	file = reader.File()

	if !bytes.Equal(reader.Sha256Hash(), file.GetSha256()) {
		return nil, ErrDigestMismatch
	}

	return file, nil
}

func EncodeVariant(id string, variant *ImageVariant) string {
	if len(id) < 64 {
		return id
	}

	bs, _ := proto.Marshal(variant)

	return id[:64] + base64.RawURLEncoding.EncodeToString(bs)
}
