package main

import (
	"context"
	"net"
	"os"

	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/error/mongoerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	logger "gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/document/internal/model/mongo"
	"gitlab.mashroom.com/mashroom/backend/document/internal/service"
	"gitlab.mashroom.com/mashroom/backend/document/internal/storage/filesystem"
	"gitlab.mashroom.com/mashroom/backend/document/version"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

var (
	envMongoURL      = env.ResolveString("MONGO_URL", "mongodb:27017")
	envMongoUsername = env.ResolveString("MONGO_USERNAME", "document-data")
	envMongoPassword = env.ResolveString("MONGO_PASSWORD", "document-data")
	envMongoAuthDB   = env.ResolveString("MONGO_AUTH", "admin")
	envMongoDB       = env.ResolveString("MONGO_DATABASE", "document-data")
	envStorageDir    = env.ResolveString("STORAGE_DIR", "/storage")
	selfRemote       = env.ResolveString("SELF_REMOTE", "mashroom-invoice:10000")

	notificationService = notification.NewNotificationServiceClient(grpc.Dial(
		env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000"),
		"Notification"))
	invoiceService = invoice.NewInvoiceServiceClient(grpc.Dial(
		env.ResolveString("INVOICE_SERVICE_URL", "mashroom-invoice:10000"),
		"Invoice"))
)

func main() {
	ctx := context.Background()

	repository, err := mongo.NewRepository(ctx, &mongo.Config{
		URL:        envMongoURL,
		Username:   envMongoUsername,
		Password:   envMongoPassword,
		Database:   envMongoDB,
		AuthSource: envMongoAuthDB,
	})
	if err != nil {
		panic(err)
	}

	err = repository.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	tempdir := envStorageDir + "/temp"

	err = os.MkdirAll(tempdir, os.FileMode(0777))
	if err != nil {
		panic(err)
	}

	storage, err := filesystem.NewRepository(ctx, &filesystem.Config{
		Directory:        envStorageDir,
		MaxOriginalWidth: 1312,
		ThumbWidth:       510,
	})
	if err != nil {
		panic(err)
	}

	go func() {
		err = storage.Migrate(ctx)
		if err != nil {
			panic(err)
		}

		err = repository.MigrateStorage(ctx, storage)
		if err != nil {
			panic(err)
		}
	}()

	connections := service.Connections{
		SelfRemote:   selfRemote,
		Notification: notificationService,
		Invoice:      invoiceService,
	}

	srv, err := service.NewService(ctx, &service.Config{
		Repository:    repository,
		Storage:       storage,
		TempDirectory: tempdir,
		Connections:   connections,
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("document", grpc.ServerOptionErrorTranslator(
		mongoerr.Translate,
	))
	if err != nil {
		panic(err)
	}

	document.RegisterDocumentServiceServer(grpcServer, srv)

	logger.Infoln("documents started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
