// Package mongo provides mongo repository implementation for document service
package mongo

import (
	"context"
	"fmt"
	"time"

	v3 "gitlab.mashroom.com/mashroom/backend/document/internal/model/mongo/migrations/v3"

	"gitlab.mashroom.com/mashroom/backend/document/internal/storage"

	v2 "gitlab.mashroom.com/mashroom/backend/document/internal/model/mongo/migrations/v2"

	document "gitlab.mashroom.com/mashroom/backend/document/api"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"

	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/generic/migrations"
	"gitlab.mashroom.com/mashroom/backend/common/generic/mongoquery"
	"gitlab.mashroom.com/mashroom/backend/document/internal/model"
	v1 "gitlab.mashroom.com/mashroom/backend/document/internal/model/mongo/migrations/v1"
)

const (
	documentServiceCollection = "documents"
)

// Config mongo repository config
type Config struct {
	URL        string
	Username   string
	Password   string
	Database   string
	AuthSource string
}

type repositoryImpl struct {
	client              *mongo.Client
	database            *mongo.Database
	collectionDocuments *mongo.Collection
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid repository config")
	}

	if config.URL == "" {
		return fmt.Errorf("empty url in repository config")
	}

	if config.Username == "" {
		return fmt.Errorf("empty username in repository config")
	}

	if config.Password == "" {
		return fmt.Errorf("empty password in repository config")
	}

	if config.Database == "" {
		return fmt.Errorf("empty database in repository config")
	}

	if config.AuthSource == "" {
		return fmt.Errorf("empty database in repository config")
	}

	return nil
}

// NewRepository returns new repository instance
func NewRepository(ctx context.Context, config *Config) (*repositoryImpl, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	client, err := mongo.NewClient(
		options.Client().ApplyURI(fmt.Sprintf("mongodb://%s", config.URL)).SetAuth(options.Credential{
			AuthSource:  config.AuthSource,
			Username:    config.Username,
			Password:    config.Password,
			PasswordSet: true,
		}),
	)
	if err != nil {
		return nil, err
	}

	err = client.Connect(ctx)
	if err != nil {
		return nil, err
	}

	db := client.Database(config.Database)

	return &repositoryImpl{
		client:              client,
		database:            db,
		collectionDocuments: db.Collection(documentServiceCollection),
	}, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	err := migrations.MongoMigrate(repo.client, repo.database.Name(),
		&migrations.MongoMigration{
			Version: "0",
			Up: func(ctx context.Context, db *mongo.Database) error {
				return nil
			},
		},
		&migrations.MongoMigration{
			Version: "1",
			Up:      v1.Up,
		},
		&migrations.MongoMigration{
			Version: "2",
			Up:      v2.Up,
		},
		&migrations.MongoMigration{
			Version: "3",
			Up:      v3.Up,
		},
	)
	if err != nil {
		return err
	}

	return nil
}

func (repo *repositoryImpl) MigrateStorage(ctx context.Context, stor storage.Repository) error {
	col := repo.database.Collection(documentServiceCollection)

	cur, err := col.Find(ctx, bson.M{"sha256": bson.M{"$in": bson.A{nil, ""}}})
	if err == mongo.ErrNoDocuments {
		return nil
	}

	if err != nil {
		return err
	}

	var m model.Document

	defer func() {
		_ = cur.Close(ctx)
	}()

	for cur.Next(ctx) {
		err = cur.Decode(&m)
		if err == mongo.ErrNoDocuments {
			continue
		}

		if err != nil {
			return err
		}

		var f *document.File

		f, _, err = stor.FileGetByIDSize(ctx, m.File, document.ImageSize_ORIGINAL, nil)
		if err != nil {
			continue
		}

		if m.File == "" {
			continue
		}

		m.Sha256 = f.GetSha256()
		m.Name = f.GetName()
		m.CreatedAt = f.GetCreatedAt().AsTime()
		m.Size = f.GetSize()
		m.Mime = f.GetMime()

		_, err = col.ReplaceOne(ctx, bson.M{"_id": m.ID}, m)
		if err != nil {
			continue
		}
	}

	return nil
}

func (repo *repositoryImpl) DocumentGetByID(
	ctx context.Context,
	id primitive.ObjectID,
) (m *model.Document, err error) {
	filter := bson.M{
		"_id": id,
	}

	m = &model.Document{}

	err = repo.collectionDocuments.FindOne(ctx, filter).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) DocumentGetByFileUserPropertyType(
	ctx context.Context,
	file, user, property string,
	typ document.DocumentType,
	removed bool,
) (m *model.Document, err error) {
	filter := bson.M{}

	if file != "" {
		filter["file"] = file
	}

	if user != "" {
		filter["user"] = user
	}

	if property != "" {
		filter["property"] = property
	}

	if typ != 0 {
		filter["type"] = typ
	}

	if !removed {
		filter["removed"] = false
	}

	m = &model.Document{}

	err = repo.collectionDocuments.FindOne(ctx, filter).Decode(m)
	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) DocumentCreate(ctx context.Context, document *model.Document) error {
	document.ID = primitive.NewObjectID()
	document.CreatedAt = time.Now()
	document.UpdatedAt = time.Now()

	_, err := repo.collectionDocuments.InsertOne(ctx, document)

	return err
}

func (repo *repositoryImpl) DocumentUpdate(ctx context.Context, document *model.Document) error {
	document.UpdatedAt = time.Now()

	_, err := repo.collectionDocuments.ReplaceOne(ctx, bson.M{"_id": document.ID}, document)

	return err
}

func (repo *repositoryImpl) DocumentRemove(ctx context.Context, id primitive.ObjectID) error {
	_, err := repo.collectionDocuments.DeleteOne(ctx, bson.M{"_id": id})
	if err != nil {
		return err
	}

	return nil
}

func (repo *repositoryImpl) DocumentCount(ctx context.Context, file string) (int64, error) {
	return repo.collectionDocuments.CountDocuments(ctx, bson.M{"file": file})
}

func (repo *repositoryImpl) DocumentFilter(
	ctx context.Context,
	params *query.QueryParams,
) (ms []*model.Document, total int64, err error) {
	filter, opts, err := mongoquery.BuildFilter(model.Document{}, params, nil)
	if err != nil {
		return
	}

	cur, err := repo.collectionDocuments.Find(ctx, filter, opts)
	if err != nil {
		return
	}

	err = cur.All(ctx, &ms)
	if err != nil {
		return
	}

	total, err = repo.collectionDocuments.CountDocuments(ctx, filter)
	if err != nil {
		return
	}

	return
}
