// Package service provides implementation for document service
package service

import (
	"context"
	"fmt"

	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/document/internal/model"
	"gitlab.mashroom.com/mashroom/backend/document/internal/storage"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
)

// Config for service implementation
type Config struct {
	// Persistent repository
	Repository model.Repository
	// File storage
	Storage storage.Repository
	// Temporary files directory
	TempDirectory string
	// Connections client connections
	Connections Connections
}

// Service implementation
type Service interface {
	document.DocumentServiceServer
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid service config")
	}

	if config.Repository == nil {
		return fmt.Errorf("no repository in service config")
	}

	if config.Storage == nil {
		return fmt.Errorf("no storage in service config")
	}

	if err := validateConfigConnections(&config.Connections); err != nil {
		return err
	}

	return nil
}

func validateConfigConnections(connections *Connections) error {
	if connections.SelfRemote == "" {
		return fmt.Errorf("no self remote address")
	}

	if connections.Notification == nil {
		return fmt.Errorf("no notification client")
	}

	if connections.Invoice == nil {
		return fmt.Errorf("no invoice client")
	}

	return nil
}

// NewService returns new service instance
func NewService(ctx context.Context, config *Config) (Service, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	srv := &serviceImpl{
		config: *config,
	}

	return srv, nil
}

// Connections remote service connections
type Connections struct {
	Notification notification.NotificationServiceClient
	Invoice      invoice.InvoiceServiceClient
	SelfRemote   string
}
