package service

import (
	"archive/zip"
	"bytes"
	"context"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"strconv"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/generic/query"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/document/internal/model"
	invoice "gitlab.mashroom.com/mashroom/backend/invoice/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"google.golang.org/protobuf/types/known/timestamppb"
)

var (
	errUnauthorized = grpcerr.New(grpcerr.AuthError, "not authorized")
)

type serviceImpl struct {
	config Config
}

func (srv *serviceImpl) tryUpdateExistingDocument(
	ctx context.Context,
	file, user, property, code, desc string,
	typ document.DocumentType,
	amount float64,
	expiry time.Time,
) (f *document.File, resp *document.CreateResponse, err error) {
	f, _, err = srv.config.Storage.FileGetByIDSize(ctx, file, document.ImageSize_ORIGINAL, nil)
	if err != nil {
		return nil, nil, err
	}

	cand, err := srv.config.Repository.DocumentGetByFileUserPropertyType(
		ctx,
		file,
		user,
		property,
		typ,
		true,
	)
	if err != nil {
		return f, nil, nil
	}

	cand.Removed = false
	cand.Expiration = expiry
	cand.Amount = amount
	cand.Code = code
	cand.Sha256 = f.GetSha256()
	cand.Name = f.GetName()
	cand.Size = f.GetSize()
	cand.Mime = f.GetMime()
	cand.State = document.DocumentState_UNAPPROVED
	cand.Description = desc

	err = srv.config.Repository.DocumentUpdate(ctx, cand)
	if err != nil {
		return f, nil, err
	}

	return nil, &document.CreateResponse{
		Success: true,
		Id:      cand.ID.Hex(),
	}, nil
}

func (srv *serviceImpl) Create(
	ctx context.Context,
	req *document.CreateRequest,
) (resp *document.CreateResponse, err error) {
	var f *document.File

	if req.GetFile() != "" {
		f, resp, err = srv.tryUpdateExistingDocument(
			ctx,
			req.GetFile(),
			req.GetUser(),
			req.GetProperty(),
			req.GetCode(),
			req.GetDescription(),
			req.GetType(),
			req.GetAmount(),
			req.GetExpiration().AsTime(),
		)
		if err != nil {
			return nil, err
		}

		if resp != nil {
			return
		}
	}

	m := &model.Document{
		Expiration:  req.GetExpiration().AsTime(),
		User:        req.GetUser(),
		Property:    req.GetProperty(),
		File:        req.GetFile(),
		Name:        f.GetName(),
		Sha256:      f.GetSha256(),
		Size:        f.GetSize(),
		Mime:        f.GetMime(),
		Description: req.GetDescription(),
		Code:        req.GetCode(),
		Amount:      req.GetAmount(),
		Type:        req.GetType(),
		Removed:     false,
		ProductID:   req.GetProductId(),
	}

	err = srv.config.Repository.DocumentCreate(ctx, m)
	if err != nil {
		return nil, err
	}

	if m.ProductID != "" && !req.GetSkipNotify() {
		err = srv.addNewDocumentInProduct(ctx, m)
		if err != nil {
			return nil, err
		}

		srv.sendNotificationDocumentAssociatedWithProduct(ctx, m)
	}

	return &document.CreateResponse{
		Success: true,
		Id:      m.ID.Hex(),
	}, nil
}

func (srv *serviceImpl) Update(ctx context.Context, req *document.UpdateRequest) (*document.UpdateResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if m.User != req.GetUser() {
		return nil, errUnauthorized
	}

	if req.GetFile() != "" {
		var f *document.File

		f, _, err = srv.config.Storage.FileGetByIDSize(ctx, req.GetFile(), document.ImageSize_ORIGINAL, nil)
		if err != nil {
			return nil, err
		}

		m.Sha256 = f.GetSha256()
		m.Name = f.GetName()
		m.Size = f.GetSize()
		m.Mime = f.GetMime()
		m.File = f.GetId()
	}

	if req.GetExpiration() != nil && req.GetExpiration().IsValid() {
		m.Expiration = req.GetExpiration().AsTime()
	}

	if req.Amount != 0 {
		m.Amount = req.GetAmount()
	}

	if req.GetCode() != "" {
		m.Code = req.GetCode()
	}

	if req.GetDescription() != "" {
		m.Description = req.GetDescription()
	}

	oldProduct := m.ProductID
	m.ProductID = req.ProductId
	m.Removed = req.GetRemoved()
	m.State = document.DocumentState_UNAPPROVED

	err = srv.config.Repository.DocumentUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	if oldProduct != m.ProductID && !req.GetSkipNotify() {
		err = srv.deleteDocumentFromProduct(ctx, m, oldProduct)
		if err != nil {
			return nil, err
		}

		if m.ProductID != "" {
			err = srv.addNewDocumentInProduct(ctx, m)
			if err != nil {
				return nil, err
			}

			srv.sendNotificationDocumentAssociatedWithProduct(ctx, m)
		}
	}

	return &document.UpdateResponse{
		Success: true,
	}, nil
}

func (srv *serviceImpl) Get(ctx context.Context, req *document.GetRequest) (*document.GetResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	doc, err := srv.renderDocument(ctx, m)
	if err != nil {
		return nil, err
	}

	return &document.GetResponse{
		Document: doc,
	}, nil
}

func (srv *serviceImpl) List(ctx context.Context, req *document.ListRequest) (*document.ListResponse, error) {
	params := &query.QueryParams{
		Pagination: &query.QueryParamsPagination{
			Page:    1,
			PerPage: 1000,
		},
		Filter: []*query.QueryParamsFilter{},
	}

	if req.GetId() != "" {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"_id"},
			Value:    []string{req.GetId()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	if req.GetFile() != "" {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"file"},
			Value:    []string{req.GetFile()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	if req.GetUser() != "" {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"user"},
			Value:    []string{req.GetUser()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	if req.GetProperty() != "" {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"property"},
			Value:    []string{req.GetProperty()},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	if len(req.GetTypes()) > 0 {
		var types []string

		for _, t := range req.GetTypes() {
			types = append(types, t.String())
		}

		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"type"},
			Value:    types,
			Operator: query.QueryFilterOperator_ONEOF,
		})
	}

	if !req.GetRemoved() {
		params.Filter = append(params.Filter, &query.QueryParamsFilter{
			Fields:   []string{"removed"},
			Value:    []string{"false"},
			Operator: query.QueryFilterOperator_EQUALS,
		})
	}

	ms, _, err := srv.config.Repository.DocumentFilter(ctx, params)
	if err != nil {
		return nil, err
	}

	docs, err := srv.renderDocuments(ctx, ms)
	if err != nil {
		return nil, err
	}

	return &document.ListResponse{Documents: docs}, nil
}

func (srv *serviceImpl) Delete(ctx context.Context, req *document.DeleteRequest) (*document.DeleteResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err == mongo.ErrNoDocuments && len(req.GetId()) == 64 {
		err = srv.config.Storage.FileDelete(ctx, req.GetId())
		if err != nil {
			return nil, err
		}

		return &document.DeleteResponse{Success: true}, nil
	}

	if err != nil {
		return nil, err
	}

	if !req.GetAdminSite() && m.User != req.GetUser() {
		return nil, errUnauthorized
	}

	err = srv.config.Repository.DocumentRemove(ctx, m.ID)
	if err != nil {
		return nil, err
	}

	var n int64

	if m.File != "" {
		n, err = srv.config.Repository.DocumentCount(ctx, m.File)
		if err != nil {
			return nil, err
		}
	}

	if n == 0 && m.File != "" {
		err = srv.config.Storage.FileDelete(ctx, m.File)
		if err != nil {
			return nil, err
		}
	}

	return &document.DeleteResponse{Success: true}, nil
}

func (srv *serviceImpl) UpdateSignStatus(
	ctx context.Context,
	req *document.UpdateSignStatusRequest,
) (*document.UpdateSignStatusResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	m.SignatureRequestID = req.GetSignatureRequestId()
	m.Signed = req.GetSigned()

	err = srv.config.Repository.DocumentUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &document.UpdateSignStatusResponse{Success: true}, nil
}

func (srv *serviceImpl) UpdateStatus(
	ctx context.Context,
	req *document.UpdateStatusRequest,
) (*document.UpdateStatusResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	m.State = req.GetState()

	err = srv.config.Repository.DocumentUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	return &document.UpdateStatusResponse{Success: true}, nil
}

func (srv *serviceImpl) bundleAddFile(ctx context.Context, w *zip.Writer, idx int, id string) error {
	file, path, err := srv.config.Storage.FileGetByIDSize(ctx, id, document.ImageSize_ORIGINAL, nil)
	if err != nil {
		return err
	}

	bs, err := ioutil.ReadFile(path)
	if err != nil {
		return err
	}

	f, err := w.Create(fmt.Sprintf("%03d_%s", idx, file.Name))
	if err != nil {
		return err
	}

	_, err = f.Write(bs)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) BundleDownload(
	req *document.BundleDownloadRequest,
	resp document.DocumentService_BundleDownloadServer,
) error {
	list, err := srv.List(resp.Context(), &document.ListRequest{
		User:     req.GetUser(),
		Property: req.GetProperty(),
		Types:    req.GetTypes(),
	})
	if err != nil {
		return err
	}

	var types []string

	for _, t := range req.GetTypes() {
		types = append(types, strconv.FormatInt(int64(t), 10))
	}

	name := req.GetProperty() + "_" + strings.Join(types, "_")
	descr := fmt.Sprintf("%s %v", req.GetProperty(), req.GetTypes())

	zipbuf := &bytes.Buffer{}

	w := zip.NewWriter(zipbuf)

	for i, m := range list.GetDocuments() {
		if m.GetFile() == nil {
			continue
		}

		_ = srv.bundleAddFile(resp.Context(), w, i, m.GetFile().GetId())
	}

	err = w.Close()
	if err != nil {
		return err
	}

	file := &document.File{
		Name:        name + ".zip",
		Mime:        "application/zip",
		CreatedAt:   timestamppb.Now(),
		Description: descr,
	}

	path, err := srv.config.Storage.FileSave(resp.Context(), file, zipbuf)
	if err != nil {
		return err
	}

	err = resp.Send(&document.FileEvent{
		Init: &document.FileInit{
			File: file,
		},
	})
	if err != nil {
		return err
	}

	f, err := os.Open(path)
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	buf := make([]byte, document.ChunkSize)

	writer := &document.UploaderWriter{
		Client: resp,
	}

	_, err = io.CopyBuffer(writer, document.PlainReader{R: f}, buf)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) FileUpload(req document.DocumentService_FileUploadServer) (err error) {
	var file *document.File

	defer func() {
		if err != nil {
			log.WithError(err).Errorf("Uploading file error: %+v", file)
		}

		log.Infof("Saved file: %+v", file)
	}()

	log.Infof("Uploading file")

	buf := make([]byte, document.ChunkSize)

	f, err := ioutil.TempFile(srv.config.TempDirectory, "download")
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
		_ = os.Remove(f.Name())
	}()

	file, err = document.DownloadFileBuffer(req.Context(), req, f, buf)
	if err != nil {
		return err
	}

	log.Infof("Uploaded file: %+v", file)

	_, err = f.Seek(0, io.SeekStart)
	if err != nil {
		return err
	}

	_, err = srv.config.Storage.FileSave(req.Context(), file, f)
	if err != nil {
		return err
	}

	return req.SendAndClose(&document.FileUploadResponse{
		Success: true,
		File:    file,
	})
}

func (srv *serviceImpl) FileDownload(
	req *document.FileDownloadRequest,
	resp document.DocumentService_FileDownloadServer,
) error {
	file, path, err := srv.config.Storage.FileGetByIDSize(resp.Context(), req.GetId(), req.GetSize(), req.GetVariant())
	if err != nil {
		return err
	}

	err = resp.Send(&document.FileEvent{
		Init: &document.FileInit{
			File: file,
		},
	})
	if err != nil {
		return err
	}

	f, err := os.Open(path)
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	buf := make([]byte, document.ChunkSize)

	writer := &document.UploaderWriter{
		Client: resp,
	}

	_, err = io.CopyBuffer(writer, document.PlainReader{R: f}, buf)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) FileUpdate(
	ctx context.Context,
	req *document.FileUpdateRequest,
) (*document.FileUpdateResponse, error) {
	err := srv.config.Storage.FileUpdate(ctx, req.GetFile())
	if err != nil {
		return nil, err
	}

	return &document.FileUpdateResponse{
		Success: true,
		File:    req.GetFile(),
	}, nil
}

func (srv *serviceImpl) FileInfo(
	ctx context.Context,
	req *document.FileInfoRequest,
) (*document.FileInfoResponse, error) {
	file, _, err := srv.config.Storage.FileGetByIDSize(ctx, req.GetId(), req.GetSize(), req.GetVariant())
	if err != nil {
		return nil, err
	}

	file.Id = req.GetId()

	return &document.FileInfoResponse{
		Success: true,
		File:    file,
	}, nil
}

func (srv *serviceImpl) ServiceCreate(
	ctx context.Context,
	req *document.ServiceCreateRequest,
) (*document.CreateResponse, error) {
	_, _, err := srv.config.Storage.FileGetByIDSize(ctx, req.GetFile(), document.ImageSize_ORIGINAL, nil)
	if err != nil {
		return nil, err
	}

	cand, err := srv.config.Repository.DocumentGetByFileUserPropertyType(
		ctx,
		req.GetFile(),
		req.GetUser(),
		req.GetProperty(),
		req.GetType(),
		true,
	)
	if err == nil {
		cand.State = req.GetState()
		cand.Expiration = req.GetExpiration().AsTime()
		cand.Amount = req.GetAmount()
		cand.Description = req.GetDescription()
		cand.Removed = req.GetRemoved()
		oldProduct := cand.ProductID
		cand.ProductID = req.GetProductId()

		err = srv.config.Repository.DocumentUpdate(ctx, cand)
		if err != nil {
			return nil, err
		}

		if oldProduct != cand.ProductID && cand.ProductID != "" && !req.GetSkipNotify() {
			srv.sendNotificationDocumentAssociatedWithProduct(ctx, cand)
		}

		return &document.CreateResponse{
			Success: true,
			Id:      cand.ID.Hex(),
		}, nil
	}

	m := &model.Document{
		ID:          primitive.NewObjectID(),
		Type:        req.GetType(),
		User:        req.GetUser(),
		Property:    req.GetProperty(),
		File:        req.GetFile(),
		Expiration:  req.GetExpiration().AsTime(),
		State:       req.GetState(),
		Code:        req.GetCode(),
		Amount:      req.GetAmount(),
		Description: req.GetDescription(),
		Removed:     req.GetRemoved(),
		ProductID:   req.GetProductId(),
	}

	err = srv.config.Repository.DocumentCreate(ctx, m)
	if err != nil {
		return nil, err
	}

	if m.ProductID != "" && !req.GetSkipNotify() {
		err = srv.addNewDocumentInProduct(ctx, m)
		if err != nil {
			return nil, err
		}

		srv.sendNotificationDocumentAssociatedWithProduct(ctx, m)
	}

	return &document.CreateResponse{
		Success: true,
		Id:      m.ID.Hex(),
	}, nil
}

func (srv *serviceImpl) ServiceUpdate(
	ctx context.Context,
	req *document.ServiceUpdateRequest,
) (*document.ServiceUpdateResponse, error) {
	id, err := primitive.ObjectIDFromHex(req.GetId())
	if err != nil {
		return nil, err
	}

	m, err := srv.config.Repository.DocumentGetByID(ctx, id)
	if err != nil {
		return nil, err
	}

	if req.GetFile() != "" {
		var f *document.File

		f, _, err = srv.config.Storage.FileGetByIDSize(ctx, req.GetFile(), document.ImageSize_ORIGINAL, nil)
		if err != nil {
			return nil, err
		}

		m.Sha256 = f.GetSha256()
		m.Name = f.GetName()
		m.Size = f.GetSize()
		m.Mime = f.GetMime()
		m.File = f.GetId()
	}

	if req.GetExpiration() != nil && req.GetExpiration().IsValid() {
		m.Expiration = req.GetExpiration().AsTime()
	}

	if req.Amount != 0 {
		m.Amount = req.GetAmount()
	}

	if req.GetCode() != "" {
		m.Code = req.GetCode()
	}

	if req.GetDescription() != "" {
		m.Description = req.GetDescription()
	}

	if req.GetUser() != "" {
		m.User = req.GetUser()
	}

	if req.GetState() != document.DocumentState_UNKNOWN_DOCUMENT_STATE {
		m.State = req.GetState()
	}

	m.Removed = req.GetRemoved()

	oldProduct := m.ProductID
	m.ProductID = req.ProductId

	err = srv.config.Repository.DocumentUpdate(ctx, m)
	if err != nil {
		return nil, err
	}

	if oldProduct != m.ProductID && !req.GetSkipNotify() {
		err = srv.deleteDocumentFromProduct(ctx, m, oldProduct)
		if err != nil {
			return nil, err
		}

		if m.ProductID != "" {
			err = srv.addNewDocumentInProduct(ctx, m)
			if err != nil {
				return nil, err
			}

			srv.sendNotificationDocumentAssociatedWithProduct(ctx, m)
		}
	}

	return &document.ServiceUpdateResponse{Success: true}, nil
}

func (srv *serviceImpl) CleanFiles(
	ctx context.Context,
	req *document.CleanFilesRequest,
) (*document.CleanFilesResponse, error) {
	err := srv.config.Storage.FileCleanSize(ctx, req.GetSize())
	if err != nil {
		return nil, err
	}

	return &document.CleanFilesResponse{}, nil
}

func (srv *serviceImpl) sendNotificationDocumentAssociatedWithProduct(ctx context.Context, doc *model.Document) {
	_, err := srv.config.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_DOCUMENT,
		Action: notification.Action_ACTION_UPDATE,
		Targets: []*notification.Target{
			{
				User: doc.User,
				Role: notification.Role_ROLE_UNKNOWN,
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_DOCUMENT,
				EmbedValue: doc.ID.Hex(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: doc.Property,
			},
		},
	})
	if err != nil {
		log.WithError(err).Info("EVENT_ADMIN_UPLOADED_DOC_NOTIFY_LANDLORD error")
	}
}

func (srv *serviceImpl) addNewDocumentInProduct(ctx context.Context, doc *model.Document) error {
	prod, err := srv.config.Connections.Invoice.ProductGet(ctx, &invoice.ProductGetRequest{Id: doc.ProductID})
	if err != nil {
		return err
	}

	prod.Product.Documents = append(prod.Product.Documents, &invoice.ProductDocument{DocumentId: doc.ID.Hex()})

	_, err = srv.config.Connections.Invoice.ProductAddDocuments(ctx, &invoice.ProductAddDocumentsRequest{
		ProductId: doc.ProductID,
		Documents: prod.Product.Documents,
		Time: &invoice.ProductDetailsTime{
			Start: timestamppb.New(doc.CreatedAt),
			End:   timestamppb.New(doc.Expiration),
		},
	})
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) deleteDocumentFromProduct(ctx context.Context, doc *model.Document, oldProduct string) error {
	prod, err := srv.config.Connections.Invoice.ProductGet(ctx, &invoice.ProductGetRequest{Id: oldProduct})
	if err != nil {
		return err
	}

	var newDocuments []*invoice.ProductDocument

	for _, d := range prod.GetProduct().GetDocuments() {
		if d.DocumentId != doc.ID.Hex() {
			newDocuments = append(newDocuments, d)
		}
	}

	_, err = srv.config.Connections.Invoice.ProductAddDocuments(ctx, &invoice.ProductAddDocumentsRequest{
		ProductId: oldProduct,
		Documents: newDocuments,
	})
	if err != nil {
		return err
	}

	return nil
}
