// Package filesystem provides file storage repository on filesystem
package filesystem

import (
	"bytes"
	"context"
	"crypto/sha256"
	"encoding/base64"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"image"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"strconv"
	"strings"
	"time"

	"github.com/chai2010/webp"
	"github.com/disintegration/imaging"
	"github.com/golang/protobuf/ptypes/timestamp"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	"gitlab.mashroom.com/mashroom/backend/document/internal/storage"
	"gitlab.mashroom.com/mashroom/backend/document/internal/storage/filesystem/migrations"
	v1 "gitlab.mashroom.com/mashroom/backend/document/internal/storage/filesystem/migrations/v1"
	"google.golang.org/protobuf/proto"
)

const (
	sha256Suffix    = ".sha256"
	jsonSuffix      = ".json"
	webpSuffix      = ".webp"
	thumbInfix      = ".thumb"
	manifestVersion = 1

	variantIVRWatermarkPrefix = ".ivr_watermark"
)

// Config filesystem storage repository config
type Config struct {
	Directory        string
	MaxOriginalWidth int
	ThumbWidth       int
}

type manifest struct {
	ID        string `json:"id"`
	Name      string `json:"name"`
	Mime      string `json:"mime"`
	Sha256    []byte `json:"sha256"`
	CreatedAt int64  `json:"created_at"`
	Size      int64  `json:"size"`
	Version   int    `json:"version"`
}

type repositoryImpl struct {
	config Config
}

func hashID(bs []byte) string {
	return hex.EncodeToString(bs)
}

func idPath(id string) (path string) {
	if len(id) > 64 {
		id = id[:64]
	}

	for len(id) > 4 {
		path, id = path+"/"+id[:4], id[4:]
	}

	if len(id) > 0 {
		path = path + "/" + id
	}

	return
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid repository config")
	}

	if config.Directory == "" {
		return fmt.Errorf("empty directory in repository config")
	}

	return nil
}

// NewRepository returns new repository instance
func NewRepository(ctx context.Context, config *Config) (storage.Repository, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	return &repositoryImpl{
		config: *config,
	}, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	err := migrations.FilesystemMigrate(repo.config.Directory,
		&migrations.Migration{
			Version: 1,
			Up:      v1.MakeUp(repo.config.Directory+"/webps", repo.config.Directory+"/thumbs"),
		},
	)
	if err != nil {
		return err
	}

	return nil
}

func readHashSize(path string) (hash []byte, size int64, err error) {
	st, err := os.Stat(path)
	if err != nil {
		return
	}

	bs, err := ioutil.ReadFile(path + sha256Suffix)
	if err != nil {
		return
	}

	n, err := hex.Decode(bs, bs)

	return bs[:n], st.Size(), err
}

func (repo *repositoryImpl) pathOrig(path, variantPath string) string {
	return path + variantPath
}

func (repo *repositoryImpl) pathThumb(path, variantPath string) string {
	return path + variantPath + thumbInfix + strconv.FormatInt(int64(repo.config.ThumbWidth), 10)
}

func (repo *repositoryImpl) pathWebP(path, variantPath string) string {
	return path + variantPath + webpSuffix
}

func (repo *repositoryImpl) variantPredicate(img image.Image, variant *document.ImageVariant) image.Image {
	if variant == nil {
		return img
	}

	if variant.GetIvrWatermark() == nil {
		return nil
	}

	nimg, err := variantIVRWatermark(img, variant.GetIvrWatermark().GetIvr())
	if err != nil {
		return img
	}

	return nimg
}

func (repo *repositoryImpl) imageSizeEnsure(
	file *document.File,
	path, mime string,
	size document.ImageSize,
	format imaging.Format,
	variant *document.ImageVariant,
) (string, string, error) {
	img, err := imaging.Open(path)
	if err != nil {
		return "", "", err
	}

	var variantPath string

	if variant.GetIvrWatermark() != nil {
		variantPath = variantIVRWatermarkPrefix + "_" + variant.GetIvrWatermark().GetIvr()
	}

	switch size {
	case document.ImageSize_ORIGINAL:
		origpath := repo.pathOrig(path, variantPath)

		file.Sha256, file.Size, err = readHashSize(origpath)
		if os.IsNotExist(err) {
			file.Sha256, file.Size, err = repo.imageProcessVariant(
				repo.variantPredicate(img, variant),
				format,
				origpath,
			)
		}

		path = origpath
	case document.ImageSize_THUMBNAIL:
		thumbpath := repo.pathThumb(path, variantPath)

		file.Sha256, file.Size, err = readHashSize(thumbpath)
		if os.IsNotExist(err) {
			file.Sha256, file.Size, err = repo.imageProcessThumbnail(
				repo.variantPredicate(img, variant),
				format,
				thumbpath,
			)
		}

		path = thumbpath
	case document.ImageSize_WEBP:
		webppath := repo.pathWebP(path, variantPath)

		file.Sha256, file.Size, err = readHashSize(webppath)
		if os.IsNotExist(err) {
			file.Sha256, file.Size, err = repo.imageProcessWebP(
				repo.variantPredicate(img, variant),
				webppath,
			)
		}

		path = webppath
		mime = "image/webp"
	}

	if err != nil {
		return "", "", err
	}

	return path, mime, nil
}

func (repo *repositoryImpl) FileGetByIDSize(
	ctx context.Context,
	id string,
	size document.ImageSize,
	variant *document.ImageVariant,
) (file *document.File, path string, err error) {
	var variantraw string

	if len(id) > 64 {
		variantraw = id[64:]
	}

	path = repo.config.Directory + idPath(id)

	jf, err := os.Open(path + jsonSuffix)
	if err != nil {
		return
	}

	defer func() {
		_ = jf.Close()
	}()

	var m manifest

	err = json.NewDecoder(jf).Decode(&m)
	if err != nil {
		return
	}

	file = &document.File{
		Id:   m.ID,
		Name: m.Name,
		Mime: m.Mime,
		CreatedAt: &timestamp.Timestamp{
			Seconds: m.CreatedAt,
		},
		Size:   m.Size,
		Sha256: m.Sha256,
	}

	if variantraw != "" && variant == nil {
		v := &document.ImageVariant{}
		bs, _ := base64.RawURLEncoding.DecodeString(variantraw)

		if len(bs) > 0 {
			err = proto.Unmarshal(bs, v)
			if err == nil {
				variant = v
			}
		}
	}

	if format := imageFormat(file.Mime); (size != document.ImageSize_ORIGINAL || variant != nil) && format != -1 {
		path, file.Mime, err = repo.imageSizeEnsure(file, path, file.Mime, size, format, variant)
		if err != nil {
			return
		}
	}

	return
}

func imageFormat(mime string) imaging.Format {
	switch mime {
	case "image/jpeg":
		return imaging.JPEG
	case "image/png":
		return imaging.PNG
	default:
		return -1
	}
}

func (repo *repositoryImpl) imageEncode(img image.Image, format imaging.Format) (bs []byte, err error) {
	buf := &bytes.Buffer{}

	err = imaging.Encode(buf, img, format)
	if err != nil {
		return nil, err
	}

	return buf.Bytes(), nil
}

func (repo *repositoryImpl) imageProcessOriginal(
	file *document.File,
	data io.Reader,
	format imaging.Format,
) (img image.Image, path string, err error) {
	img, err = imaging.Decode(data, imaging.AutoOrientation(true))
	if err != nil {
		return
	}

	if img.Bounds().Size().X > repo.config.MaxOriginalWidth {
		img = imaging.Resize(img, repo.config.MaxOriginalWidth, 0, imaging.NearestNeighbor)
	}

	bs, err := repo.imageEncode(img, format)
	if err != nil {
		return
	}

	path, err = repo.fileProcess(file, bs)
	if err != nil {
		return
	}

	return
}

func (repo *repositoryImpl) imageProcessVariant(
	img image.Image,
	format imaging.Format,
	origpath string,
) (hash []byte, size int64, err error) {
	bs, err := repo.imageEncode(img, format)
	if err != nil {
		return
	}

	f, err := os.Create(origpath)
	if err != nil {
		return
	}

	defer func() {
		_ = f.Close()
	}()

	_, err = io.Copy(f, bytes.NewReader(bs))
	if err != nil {
		return
	}

	hf, err := os.Create(origpath + sha256Suffix)
	if err != nil {
		return
	}

	defer func() {
		_ = hf.Close()
	}()

	h := sha256.Sum256(bs)

	_, err = io.Copy(hf, strings.NewReader(hashID(h[:])))
	if err != nil {
		return
	}

	return h[:], int64(len(bs)), nil
}

func (repo *repositoryImpl) imageProcessThumbnail(
	img image.Image,
	format imaging.Format,
	thumbpath string,
) (hash []byte, size int64, err error) {
	if img.Bounds().Size().X > repo.config.ThumbWidth {
		img = imaging.Resize(img, repo.config.ThumbWidth, 0, imaging.NearestNeighbor)
	}

	return repo.imageProcessVariant(img, format, thumbpath)
}

func (repo *repositoryImpl) imageProcessWebP(
	img image.Image,
	webppath string,
) (hash []byte, size int64, err error) {
	buf := &bytes.Buffer{}

	err = webp.Encode(buf, img, &webp.Options{Lossless: false, Quality: 25})
	if err != nil {
		return
	}

	h := sha256.Sum256(buf.Bytes())

	f, err := os.Create(webppath)
	if err != nil {
		return
	}

	defer func() {
		_ = f.Close()
	}()

	_, err = io.Copy(f, buf)
	if err != nil {
		return
	}

	hf, err := os.Create(webppath + sha256Suffix)
	if err != nil {
		return
	}

	defer func() {
		_ = hf.Close()
	}()

	_, err = io.Copy(hf, strings.NewReader(hashID(h[:])))
	if err != nil {
		return
	}

	return h[:], int64(buf.Len()), nil
}

func (repo *repositoryImpl) imageProcess(
	file *document.File,
	data io.Reader,
	format imaging.Format,
) (path string, err error) {
	img, path, err := repo.imageProcessOriginal(file, data, format)
	if err != nil {
		return "", err
	}

	_, _, err = repo.imageProcessThumbnail(img, format, repo.pathThumb(path, ""))
	if err != nil {
		return "", err
	}

	_, _, err = repo.imageProcessWebP(img, repo.pathWebP(path, ""))
	if err != nil {
		return "", err
	}

	return path, nil
}

func (repo *repositoryImpl) fileProcess(file *document.File, bs []byte) (path string, err error) {
	hash := sha256.Sum256(bs)

	file.Id = hashID(hash[:])
	file.Sha256 = hash[:]
	file.Size = int64(len(bs))

	mbs, err := json.Marshal(&manifest{
		ID:        file.Id,
		Name:      file.Name,
		Mime:      file.Mime,
		CreatedAt: time.Now().Unix(),
		Size:      int64(len(bs)),
		Sha256:    hash[:],
		Version:   manifestVersion,
	})
	if err != nil {
		return
	}

	path = repo.config.Directory + idPath(file.Id)

	err = os.MkdirAll(filepath.Dir(path), os.FileMode(0777))
	if err != nil {
		return
	}

	f, err := os.Create(path)
	if err != nil {
		return
	}

	defer func() {
		_ = f.Sync()
		_ = f.Close()
	}()

	_, err = io.Copy(f, bytes.NewReader(bs))
	if err != nil {
		return
	}

	jf, err := os.Create(path + jsonSuffix)
	if err != nil {
		return
	}

	defer func() {
		_ = jf.Sync()
		_ = jf.Close()
	}()

	_, err = io.Copy(jf, bytes.NewReader(mbs))
	if err != nil {
		return
	}

	return
}

func (repo *repositoryImpl) FileSave(
	ctx context.Context,
	file *document.File,
	data io.Reader,
) (path string, err error) {
	if f := imageFormat(file.GetMime()); f != -1 {
		return repo.imageProcess(file, data, f)
	}

	bs, err := ioutil.ReadAll(data)
	if err != nil {
		return "", err
	}

	return repo.fileProcess(file, bs)
}

func (repo *repositoryImpl) FileUpdate(ctx context.Context, file *document.File) (err error) {
	path := repo.config.Directory + idPath(file.GetId())

	jf, err := os.OpenFile(path+jsonSuffix, os.O_RDWR|os.O_CREATE, os.FileMode(0666))
	if err != nil {
		return
	}

	defer func() {
		_ = jf.Close()
	}()

	var m manifest

	err = json.NewDecoder(jf).Decode(&m)
	if err != nil {
		return
	}

	file.Sha256 = m.Sha256
	file.Size = m.Size
	file.CreatedAt = &timestamp.Timestamp{
		Seconds: m.CreatedAt,
	}
	file.Mime = m.Mime

	m.Name = file.GetName()

	_, err = jf.Seek(0, io.SeekStart)
	if err != nil {
		return
	}

	err = jf.Truncate(0)
	if err != nil {
		return
	}

	err = json.NewEncoder(jf).Encode(m)
	if err != nil {
		return
	}

	return nil
}

func (repo *repositoryImpl) FileDelete(ctx context.Context, id string) (err error) {
	path := repo.config.Directory + idPath(id)

	paths, err := filepath.Glob(path + "*")
	if err != nil {
		return err
	}

	for _, p := range paths {
		err = os.Remove(p)
		if err != nil {
			return err
		}
	}

	return nil
}

func (repo *repositoryImpl) FileCleanSize(ctx context.Context, size document.ImageSize) (err error) {
	var suffix string

	switch size {
	case document.ImageSize_THUMBNAIL:
		suffix = thumbInfix + strconv.FormatInt(int64(repo.config.ThumbWidth), 10)
	case document.ImageSize_WEBP:
		suffix = webpSuffix
	default:
		return nil
	}

	return filepath.Walk(repo.config.Directory, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		if info.IsDir() || (!strings.HasSuffix(path, suffix) && !strings.HasSuffix(path, suffix+sha256Suffix)) {
			return nil
		}

		return os.Remove(path)
	})
}
