// Package migrations provides migrator implementation and migrations for filesystem storage
package migrations

import (
	"bytes"
	"encoding/json"
	"errors"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"sort"
	"strconv"
	"strings"

	"github.com/golang-migrate/migrate/v4"

	"github.com/golang-migrate/migrate/v4/database"
	"github.com/golang-migrate/migrate/v4/source"
)

func init() {
	database.Register("filesystem-runtime", &filesystemDriver{})
	source.Register("filesystem-runtime", &filesystemConfig{})
}

const (
	migrationLock   = "/migration.lock"
	schemaMigration = "/schema_migration.json"
)

var (
	errInvalidMigration = errors.New("invalid migration")
)

// FilesystemMigrate executes given migrations on given directory
func FilesystemMigrate(dir string, migrations ...*Migration) error {
	config := WithConfig(migrations...)

	driver := WithInstance(dir, migrations...)

	migration, err := migrate.NewWithInstance("filesystem-imperative", config, "filesystem-imperative", driver)
	if err != nil {
		return err
	}

	err = migration.Up()

	switch err {
	case nil, migrate.ErrNoChange:
		return nil
	default:
		return err
	}
}

type schemaVersion struct {
	Version int  `json:"version"`
	Dirty   bool `json:"dirty"`
}

// Migration represents single migration step
type Migration struct {
	Up      func(dir string) error
	Down    func(dir string) error
	Version uint
}

// WithInstance provides migration driver instance
func WithInstance(path string, migrations ...*Migration) database.Driver {
	return &filesystemDriver{
		dir:        path,
		migrations: migrations,
	}
}

type filesystemDriver struct {
	dir        string
	migrations []*Migration
}

func (fs *filesystemDriver) Open(url string) (database.Driver, error) {
	return &filesystemDriver{
		dir: url,
	}, nil
}

func (fs *filesystemDriver) Close() error {
	return nil
}

func (fs *filesystemDriver) Lock() error {
	lock := fs.dir + migrationLock

	_, err := os.Stat(lock)
	if !os.IsNotExist(err) {
		return database.ErrLocked
	}

	f, err := os.Create(lock)
	if err != nil {
		return err
	}

	_ = f.Close()

	return nil
}

func (fs *filesystemDriver) Unlock() error {
	return os.Remove(fs.dir + migrationLock)
}

func (fs *filesystemDriver) Run(migration io.Reader) error {
	bs, err := ioutil.ReadAll(migration)
	if err != nil {
		return err
	}

	parts := strings.Split(string(bs), "-")
	if len(parts) != 2 {
		return errInvalidMigration
	}

	id, err := strconv.ParseUint(parts[1], 10, 64)
	if err != nil {
		return err
	}

	idx := sort.Search(len(fs.migrations), func(i int) bool {
		return fs.migrations[i].Version == uint(id)
	})

	if idx == len(fs.migrations) {
		return errInvalidMigration
	}

	switch parts[0] {
	case "up":
		return fs.migrations[idx].Up(fs.dir)
	case "down":
		return fs.migrations[idx].Down(fs.dir)
	default:
		return errInvalidMigration
	}
}

func (fs *filesystemDriver) SetVersion(version int, dirty bool) error {
	bs, err := json.Marshal(schemaVersion{
		Version: version,
		Dirty:   dirty,
	})

	if err != nil {
		return err
	}

	f, err := os.Create(fs.dir + schemaMigration)
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	_, err = io.Copy(f, bytes.NewReader(bs))
	if err != nil {
		return err
	}

	return nil
}

func (fs *filesystemDriver) Version() (version int, dirty bool, err error) {
	f, err := os.Open(fs.dir + schemaMigration)
	if os.IsNotExist(err) {
		return -1, false, nil
	}

	if err != nil {
		return 0, false, err
	}

	defer func() {
		_ = f.Close()
	}()

	var v schemaVersion

	err = json.NewDecoder(f).Decode(&v)
	if err != nil {
		return 0, false, err
	}

	return v.Version, v.Dirty, nil
}

func (fs *filesystemDriver) Drop() error {
	return nil
}

// WithConfig provides config driver instance
func WithConfig(migrations ...*Migration) source.Driver {
	return &filesystemConfig{
		migrations: migrations,
	}
}

type filesystemConfig struct {
	migrations []*Migration
}

func (fs *filesystemConfig) Open(url string) (source.Driver, error) {
	return fs, nil
}

func (fs *filesystemConfig) Close() error {
	return nil
}

func (fs *filesystemConfig) First() (version uint, err error) {
	if len(fs.migrations) == 0 {
		return 0, os.ErrNotExist
	}

	return fs.migrations[0].Version, nil
}

func (fs *filesystemConfig) Prev(version uint) (prevVersion uint, err error) {
	idx := sort.Search(len(fs.migrations), func(i int) bool {
		return version >= fs.migrations[i].Version
	})

	if idx == 0 {
		return 0, os.ErrNotExist
	}

	return fs.migrations[idx-1].Version, nil
}

func (fs *filesystemConfig) Next(version uint) (prevVersion uint, err error) {
	idx := sort.Search(len(fs.migrations), func(i int) bool {
		return fs.migrations[i].Version > version
	})

	if idx == len(fs.migrations) {
		return 0, os.ErrNotExist
	}

	return fs.migrations[idx].Version, nil
}

type closerReader struct {
	io.Reader
}

func (closerReader) Close() error {
	return nil
}

func (fs *filesystemConfig) ReadUp(version uint) (r io.ReadCloser, identifier string, err error) {
	idx := sort.Search(len(fs.migrations), func(i int) bool {
		return fs.migrations[i].Version == version
	})

	if idx == len(fs.migrations) {
		return nil, "", os.ErrNotExist
	}

	ver := fmt.Sprintf("up-%d", fs.migrations[idx].Version)

	return closerReader{Reader: strings.NewReader(ver)}, ver, nil
}

func (fs *filesystemConfig) ReadDown(version uint) (r io.ReadCloser, identifier string, err error) {
	idx := sort.Search(len(fs.migrations), func(i int) bool {
		return fs.migrations[i].Version == version
	})

	if idx == len(fs.migrations) {
		return nil, "", os.ErrNotExist
	}

	ver := fmt.Sprintf("down-%d", fs.migrations[idx].Version)

	return closerReader{Reader: strings.NewReader(ver)}, ver, nil
}
