// Package v1 provides migration to version 1 of filesystem storage
package v1

import (
	"bytes"
	"crypto/sha256"
	"encoding/hex"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"os"
	"path/filepath"
	"strings"
)

type oldManifest struct {
	ID          string `json:"Id"`
	Name        string `json:"Name"`
	Mime        string `json:"Mime"`
	Description string `json:"Description"`
	Version     string `json:"Version"`
	Sha256      []byte `json:"Sha256"`
	CreatedAt   int64  `json:"CreatedAt"`
	Size        int64  `json:"Size"`
	DeletedOn   int64  `json:"DeletedOn"`
}

type newManifest struct {
	ID          string `json:"id"`
	Name        string `json:"name"`
	Mime        string `json:"mime"`
	Description string `json:"description"`
	Sha256      []byte `json:"sha256"`
	CreatedAt   int64  `json:"created_at"`
	Size        int64  `json:"size"`
	Version     int    `json:"version"`
}

func hash2Filepath(bs []byte) (path string) {
	for i := 0; i < len(bs)/2; i++ {
		path += fmt.Sprintf("/%02x%02x", bs[i*2], bs[i*2+1])
	}

	return
}

func updateManifest(path string) (n *newManifest, err error) {
	f, err := os.OpenFile(path, os.O_RDWR|os.O_CREATE, os.FileMode(0666))
	if err != nil {
		return
	}

	defer func() {
		_ = f.Close()
	}()

	bs, err := ioutil.ReadAll(f)
	if err != nil {
		return
	}

	var natt newManifest

	err = json.Unmarshal(bs, &natt)
	if err == nil && natt.Version >= 1 {
		return &natt, nil
	}

	m := &oldManifest{}

	err = json.Unmarshal(bs, m)
	if err != nil {
		return
	}

	n = &newManifest{
		ID:          m.ID,
		Name:        m.Name,
		Mime:        m.Mime,
		CreatedAt:   m.CreatedAt,
		Size:        m.Size,
		Sha256:      m.Sha256,
		Description: m.Description,
		Version:     1,
	}

	bs, err = json.Marshal(n)
	if err != nil {
		return
	}

	_, err = f.Seek(0, io.SeekStart)
	if err != nil {
		return
	}

	err = f.Truncate(0)
	if err != nil {
		return
	}

	_, err = io.Copy(f, bytes.NewReader(bs))
	if err != nil {
		return
	}

	return
}

func fileExists(path string) bool {
	_, err := os.Stat(path)

	return err == nil
}

func moveDependent(m *newManifest, dir, path, suffix string) error {
	base := dir + hash2Filepath(m.Sha256)

	target := base + suffix

	err := os.Rename(path, target)
	if err != nil {
		return err
	}

	f, err := os.Open(target)
	if err != nil {
		return err
	}

	defer func() {
		_ = f.Close()
	}()

	hash := sha256.New()

	_, err = io.Copy(hash, f)
	if err != nil {
		return err
	}

	sum := hex.EncodeToString(hash.Sum(nil))

	hashf, err := os.Create(target + ".sha256")
	if err != nil {
		return err
	}

	defer func() {
		_ = hashf.Close()
	}()

	_, err = io.Copy(hashf, strings.NewReader(sum))
	if err != nil {
		return err
	}

	return nil
}

func migrateFiles(webpdir, thumbdir, dir string) error {
	return filepath.Walk(dir, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		if info.IsDir() || !strings.HasSuffix(path, ".json") || len(filepath.Base(path)) != 9 ||
			strings.HasPrefix(path, webpdir) || strings.HasPrefix(path, thumbdir) {
			return nil
		}

		fmt.Println("migrating", path)

		m, err := updateManifest(path)
		if err != nil {
			return err
		}

		webpfile := webpdir + hash2Filepath(m.Sha256)
		thumbfile := thumbdir + hash2Filepath(m.Sha256)

		for _, v := range []string{"", ".webp"} {
			if fileExists(webpfile + v) {
				err = moveDependent(m, dir, webpfile+v, ".webp")
				if err != nil {
					return err
				}

				break
			}
		}

		for _, v := range []string{"", ".jpeg", ".jpg", ".png"} {
			if fileExists(thumbfile + v) {
				err = moveDependent(m, dir, thumbfile+v, ".thumb510")
				if err != nil {
					return err
				}

				break
			}
		}

		return nil
	})
}

// MakeUp constructs v1 Up migration
func MakeUp(webpdir, thumbdir string) func(string) error {
	return func(dir string) error {
		err := migrateFiles(webpdir, thumbdir, dir)
		if err != nil {
			return err
		}

		err = os.RemoveAll(webpdir)
		if err != nil {
			return err
		}

		err = os.RemoveAll(thumbdir)
		if err != nil {
			return err
		}

		return nil
	}
}
