package filesystem

import (
	"bytes"
	_ "embed" // embed assets
	"image"
	"image/draw"
	"image/png"

	"github.com/disintegration/imaging"
	"github.com/golang/freetype"
	"golang.org/x/image/font"
)

//go:embed assets/arial.ttf
var assetFontArial []byte

//go:embed assets/ivrtag.png
var assetIVRTag []byte

const (
	ivrFontSize    = 44.0
	ivrFontSpacing = 1.0
	ivrFontDPI     = 72.0
)

func variantIVRWatermark(img image.Image, ivr string) (image.Image, error) {
	if len(ivr) != 6 {
		return img, nil
	}

	f, err := freetype.ParseFont(assetFontArial)
	if err != nil {
		return nil, err
	}

	imgTag, _ := png.Decode(bytes.NewReader(assetIVRTag))

	fg, bg := image.White, image.Transparent

	rgba := image.NewRGBA(
		image.Rect(
			0,
			0,
			int(float64(imgTag.Bounds().Size().X)*0.6),
			int(float64(imgTag.Bounds().Size().Y)*0.4),
		),
	)

	draw.Draw(rgba, rgba.Bounds(), bg, image.Point{}, draw.Src)

	c := freetype.NewContext()

	c.SetDPI(ivrFontDPI)
	c.SetFont(f)
	c.SetFontSize(ivrFontSize)
	c.SetClip(rgba.Bounds())
	c.SetDst(rgba)
	c.SetSrc(fg)
	c.SetHinting(font.HintingFull)

	pt := freetype.Pt(0, int(c.PointToFixed(ivrFontSize)>>6))

	_, err = c.DrawString("MASH#"+ivr, pt)
	if err != nil {
		return nil, err
	}

	pt.Y += c.PointToFixed(ivrFontSize * ivrFontSpacing)

	width := float64(imgTag.Bounds().Size().X)

	height := float64(imgTag.Bounds().Size().Y)

	offset := image.Pt(int(width/3.0), int((height-float64(rgba.Bounds().Size().Y))/2.0))

	// labeled tag
	src := image.NewRGBA(imgTag.Bounds())

	draw.Draw(src, imgTag.Bounds(), imgTag, image.Point{}, draw.Src)

	draw.Draw(src, rgba.Bounds().Add(offset), rgba, image.Point{}, draw.Over)

	// main pic size
	picHeight := img.Bounds().Size().Y

	picWidth := img.Bounds().Size().X

	// tag size
	tagHeight := float64(src.Bounds().Size().Y)

	tagWidth := float64(src.Bounds().Size().X)

	// scale ratio

	// small main pic
	if picHeight < int(tagHeight) || picWidth < int(tagWidth) {
		if picHeight > picWidth {
			tagHeight = float64(picHeight) / 5.0
			tagWidth = (tagHeight / float64(src.Bounds().Size().Y)) * float64(src.Bounds().Size().X)
		} else {
			tagWidth = float64(picWidth) / 2.0
			tagHeight = (tagWidth / float64(src.Bounds().Size().X)) * float64(src.Bounds().Size().Y)
		}
	} else {
		if picHeight > picWidth {
			tagHeight = float64(picHeight) / 8.0
			tagWidth = (tagHeight / float64(src.Bounds().Size().Y)) * float64(src.Bounds().Size().X)
		} else {
			tagWidth = float64(picWidth) / 4.0
			tagHeight = (tagWidth / float64(src.Bounds().Size().X)) * float64(src.Bounds().Size().Y)
		}
	}

	watermarked := imaging.Resize(src, int(tagWidth), int(tagHeight), imaging.Linear)

	offsetPoint := image.Pt(
		int(float64(img.Bounds().Size().X-watermarked.Bounds().Size().X)*0.5),
		int(float64(img.Bounds().Size().Y)-2.0*float64(watermarked.Bounds().Size().Y)),
	)

	picBounds := img.Bounds()

	newPic := image.NewNRGBA(picBounds)

	draw.Draw(newPic, picBounds, img, image.Point{}, draw.Src)
	draw.Draw(newPic, watermarked.Bounds().Add(offsetPoint), watermarked, image.Point{}, draw.Over)

	return newPic, nil
}
